@@ -51,6 +51,7 @@ static short      Ignore_int_decl;
 static short      int_decl;
 static short      Ignore_sec_decl;
 static short      sos_decl;
+static short      Ignore_free_decl;
 static int        Rows;
 static int        Columns;
 static int        Non_zeros;
@@ -65,6 +66,7 @@ static char       *title;
 
 struct structSOSvars {
   char                 *name;
+  int                  col;
   REAL                 weight;
   struct structSOSvars *next;
 };
@@ -78,9 +80,21 @@ static struct structSOS {
   struct structSOS     *next;
 } *FirstSOS, *LastSOS;
 
+struct SOSrow {
+  int  col;
+  REAL value;
+  struct SOSrow *next;
+};
+
+struct SOSrowdata {
+  short type;
+  char *name;
+  struct SOSrow *SOSrow;
+};
+
 static struct _tmp_store_struct
 {
-  nstring name;
+  char    *name;
   int     row;
   REAL    value;
   REAL    rhs_value;
@@ -96,6 +110,7 @@ static struct rside /* contains relational operator and rhs value */
   short         relat;
   short         range_relat;
   char		negate;
+  short         SOStype;
 } *First_rside, *rs;
 
 struct column
@@ -109,6 +124,7 @@ struct column
 static struct structcoldata {
   int               must_be_int;
   int               must_be_sec;
+  int               must_be_free;
   REAL              upbo;
   REAL              lowbo;
   struct  column   *firstcol;
@@ -143,14 +159,26 @@ void add_row()
   Lin_term_count = 0;
 }
 
-void check_int_sec_sos_decl(int within_int_decl, int within_sec_decl, int sos_decl0)
+void add_sos_row(short SOStype)
+{
+  if (rs != NULL)
+    rs->SOStype = SOStype;
+  Rows++;
+  rs = NULL;
+  Lin_term_count = 0;
+}
+
+void check_int_sec_sos_free_decl(int within_int_decl, int within_sec_decl, int sos_decl0, int within_free_decl)
 {
   Ignore_int_decl = TRUE;
   Ignore_sec_decl = TRUE;
+  Ignore_free_decl = TRUE;
   sos_decl = 0;
   if(within_int_decl) {
     Ignore_int_decl = FALSE;
     int_decl = (short) within_int_decl;
+    if(within_sec_decl)
+      Ignore_sec_decl = FALSE;
   }
   else if(within_sec_decl) {
     Ignore_sec_decl = FALSE;
@@ -158,6 +186,9 @@ void check_int_sec_sos_decl(int within_int_decl, int within_sec_decl, int sos_de
   else if(sos_decl0) {
     sos_decl = (short) sos_decl0;
   }
+  else if(within_free_decl) {
+    Ignore_free_decl = FALSE;
+  }
 }
 
 static void add_int_var(char *name, short int_decl)
@@ -194,6 +225,10 @@ static void add_int_var(char *name, short int_decl)
       }
       coldata[hp->index].upbo = 1;
     }
+    else if(int_decl == 3) {
+      if(coldata[hp->index].upbo == DEF_INFINITE)
+        coldata[hp->index].upbo = 1.0;
+    }
   }
 }
 
@@ -217,6 +252,51 @@ static void add_sec_var(char *name)
     coldata[hp->index].must_be_sec = TRUE;
 }
 
+int set_sec_threshold(char *name, REAL threshold)
+{
+  hashelem *hp;
+
+  if((hp = findhash(name, Hash_tab)) == NULL) {
+    char buf[256];
+
+    sprintf(buf, "Unknown variable %s declared semi-continuous, ignored", name);
+    error(NORMAL, buf);
+    return(FALSE);
+  }
+
+  if ((coldata[hp->index].lowbo > 0.0) && (threshold > 0.0)) {
+    char buf[256];
+
+    coldata[hp->index].must_be_sec = FALSE;
+    sprintf(buf, "Variable %s declared semi-continuous, but it has a non-negative lower bound (%f), ignored", name, coldata[hp->index].lowbo);
+    error(NORMAL, buf);
+  }
+  if (threshold > coldata[hp->index].lowbo)
+    coldata[hp->index].lowbo = threshold;
+
+  return(coldata[hp->index].must_be_sec);
+}
+
+static void add_free_var(char *name)
+{
+  hashelem *hp;
+
+  if((hp = findhash(name, Hash_tab)) == NULL) {
+    char buf[256];
+
+    sprintf(buf, "Unknown variable %s declared free, ignored", name);
+    error(NORMAL, buf);
+  }
+  else if(coldata[hp->index].must_be_free) {
+    char buf[256];
+
+    sprintf(buf, "Variable %s declared free more than once, ignored", name);
+    error(NORMAL, buf);
+  }
+  else
+    coldata[hp->index].must_be_free = TRUE;
+}
+
 static int add_sos_name(char *name)
 {
   struct structSOS *SOS;
@@ -270,14 +350,19 @@ static int add_sos_var(char *name)
 
 void storevarandweight(char *name)
 {
-  if(!Ignore_int_decl)
+  if(!Ignore_int_decl) {
     add_int_var(name, int_decl);
+    if(!Ignore_sec_decl)
+      add_sec_var(name);
+  }
   else if(!Ignore_sec_decl)
     add_sec_var(name);
   else if(sos_decl==1)
     add_sos_name(name);
   else if(sos_decl==2)
     add_sos_var(name);
+  else if(!Ignore_free_decl)
+    add_free_var(name);
 }
 
 int set_sos_type(int SOStype)
@@ -287,12 +372,13 @@ int set_sos_type(int SOStype)
   return(TRUE);
 }
 
-int set_sos_weight(int weight, int sos_decl)
+int set_sos_weight(double weight, int sos_decl)
 {
-  if(sos_decl==1)
-    LastSOS->weight = weight;
-  else
-    LastSOS->LastSOSvars->weight = weight;
+  if(LastSOS != NULL)
+    if(sos_decl==1)
+      LastSOS->weight = (int) (weight+.1);
+    else
+      LastSOS->LastSOSvars->weight = weight;
   return(TRUE);
 }
 
@@ -301,7 +387,7 @@ void set_obj_dir(int maximise)
   Maximise = (short) maximise;
 }
 
-static int inccoldata()
+static int inccoldata(void)
 {
   if(Columns == 0)
     CALLOC(coldata, coldatastep, struct structcoldata);
@@ -315,6 +401,7 @@ static int inccoldata()
     coldata[Columns].firstcol = NULL;
     coldata[Columns].must_be_int = FALSE;
     coldata[Columns].must_be_sec = FALSE;
+    coldata[Columns].must_be_free = FALSE;
   }
 
   return(coldata != NULL);
@@ -340,6 +427,7 @@ static int init_read(int verbose)
     /* first row (nr 0) is always the objective function */
     rs->relat = OF;
     rs->range_relat = -1;
+    rs->SOStype = 0;
     Hash_tab = NULL;
     Hash_constraints = NULL;
     if (((Hash_tab = create_hash_table(HASHSIZE, 0)) == NULL) ||
@@ -361,7 +449,7 @@ void null_tmp_store(int init_Lin_term_count)
 {
   tmp_store.value = 0;
   tmp_store.rhs_value = 0;
-  tmp_store.name[0] = 0;
+  FREE(tmp_store.name);
   if(init_Lin_term_count)
     Lin_term_count = 0;
 }
@@ -428,7 +516,7 @@ static int store(char *variable,
   return(TRUE);
 } /* store */
 
-static int storefirst()
+static int storefirst(void)
 {
     struct rside *rp;
 
@@ -444,16 +532,19 @@ static int storefirst()
     rs->value = tmp_store.rhs_value;
     rs->relat = tmp_store.relat;
     rs->range_relat = -1;
+    rs->SOStype = 0;
 
-    if(tmp_store.value != 0) {
-      if (!store(tmp_store.name, tmp_store.row, tmp_store.value))
-        return(FALSE);
-    }
-    else {
-      char buf[256];
+    if(tmp_store.name != NULL) {
+      if(tmp_store.value != 0) {
+	if (!store(tmp_store.name, tmp_store.row, tmp_store.value))
+	  return(FALSE);
+      }
+      else {
+	char buf[256];
 
-      sprintf(buf, "Warning, variable %s has an effective coefficient of 0, ignored", tmp_store.name);
-      error(NORMAL, buf);
+	sprintf(buf, "Warning, variable %s has an effective coefficient of 0, ignored", tmp_store.name);
+	error(NORMAL, buf);
+      }
     }
     null_tmp_store(FALSE);
     return(TRUE);
@@ -482,6 +573,13 @@ int store_re_op(char *yytext, int HadConstraint, int HadVar, int Had_lineair_sum
     tmp_relat = LE;
     break;
 
+  case 0:
+    if(rs != NULL)
+      tmp_relat = rs->relat;
+    else
+      tmp_relat = tmp_store.relat;
+    break;
+
   default:
     {
       char buf[256];
@@ -494,7 +592,7 @@ int store_re_op(char *yytext, int HadConstraint, int HadVar, int Had_lineair_sum
   }
 
   if(/* Lin_term_count > 1 */ HadConstraint && HadVar) {/* it is not a bound */
-    if(Lin_term_count == 1)
+    if(Lin_term_count <= 1)
       if(!storefirst())
         return(FALSE);
     rs->relat = tmp_relat;
@@ -599,18 +697,11 @@ int var_store(char *var, REAL value, int HadConstraint, int HadVar, int Had_line
   int row;
 
   row = Rows;
-  if(strlen(var) > MAXSTRL) {
-    char buf[256];
 
-    sprintf(buf, "Variable name '%-.*s' too long, at most %d characters allowed",
-	    MAXSTRL, var, MAXSTRL);
-    error(CRITICAL, buf);
-    return(FALSE);
-  }
   /* also in a bound the same var name can occur more than once. Check for
      this. Don't increment Lin_term_count */
 
-  if(Lin_term_count != 1 || strcmp(tmp_store.name, var) != 0)
+  if(Lin_term_count != 1 || tmp_store.name == NULL || strcmp(tmp_store.name, var) != 0)
     Lin_term_count++;
 
   /* always store objective function with rownr == 0. */
@@ -618,7 +709,8 @@ int var_store(char *var, REAL value, int HadConstraint, int HadVar, int Had_line
     return(store(var,  row,  value));
 
   if(Lin_term_count == 1) { /* don't store yet. could be a bound */
-    strcpy(tmp_store.name, var);
+    if(MALLOC(tmp_store.name, strlen(var) + 1, char) != NULL)
+      strcpy(tmp_store.name, var);
     tmp_store.row = row;
     tmp_store.value += value;
     return(TRUE);
@@ -701,7 +793,7 @@ int store_bounds(int warn)
   else /* tmp_store.value = 0 ! */ {
     char buf[256];
 
-    if((tmp_store.rhs_value == 0) || 
+    if((tmp_store.rhs_value == 0) ||
        ((tmp_store.rhs_value > 0) && (tmp_store.relat == LE)) ||
        ((tmp_store.rhs_value < 0) && (tmp_store.relat == GE))) {
       sprintf(buf, "Variable %s has an effective coefficient of 0 in bound, ignored",
@@ -734,15 +826,6 @@ int add_constraint_name(char *name)
   int row;
   hashelem *hp;
 
-  if(strlen(name) > MAXSTRL) {
-    char buf[256];
-
-    sprintf(buf, "Constraint name '%-.*s' too long, at most %d characters allowed",
-	    MAXSTRL, name, MAXSTRL);
-    error(CRITICAL, buf);
-    return(FALSE);
-  }
-
   if((hp = findhash(name, Hash_constraints)) != NULL) {
     row = hp->index;
     rs = First_rside;
@@ -753,7 +836,8 @@ int add_constraint_name(char *name)
     row = Rows;
     if (((hp = puthash(name, row, NULL, Hash_constraints)) == NULL)
        ) return(FALSE);
-    rs = NULL;
+    if(row)
+      rs = NULL;
   }
 
   return(TRUE);
@@ -765,75 +849,92 @@ int add_constraint_name(char *name)
  */
 static int readinput(lprec *lp)
 {
-  int    i, count, index;
+  int    i, i1, count, index, col;
   struct column *cp, *tcp;
   hashelem *hp;
   struct rside *rp;
-  MYBOOL *negate = NULL;
+  signed char *negateAndSOS = NULL;
   REAL *row = NULL, a;
   int *rowno = NULL;
+  MYBOOL SOSinMatrix = FALSE;
+  struct SOSrowdata *SOSrowdata = NULL;
+  struct SOSrow *SOSrow, *SOSrow1;
 
   if(lp != NULL) {
-    if (CALLOC(negate, 1 + Rows, MYBOOL) == NULL)
+    if (CALLOC(negateAndSOS, 1 + Rows, signed char) == NULL)
       return(FALSE);
 
+    rp = First_rside;
+    for(i = Rows; (i >= 0) && (rp != NULL); i--) {
+      if(rp->SOStype == 0)
+        negateAndSOS[i] = (rp->negate ? -1 : 0);
+      else
+        negateAndSOS[i] = (signed char) rp->SOStype;
+
+      rp = rp->next;
+    }
+
     /* fill names with the rownames */
     hp = Hash_constraints->first;
     while(hp != NULL) {
-      if (!set_row_name(lp, hp->index, hp->name))
+      if (/* (negateAndSOS[hp->index] <= 0) && */ (!set_row_name(lp, hp->index, hp->name)))
         return(FALSE);
       hp = hp->nextelem;
     }
   }
 
-  for(i = Rows;i >= 0;i--) {
+  for(i = Rows; i >= 0; i--) {
     rp = First_rside;
     if((lp != NULL) && (rp != NULL)) {
-      negate[i] = rp->negate;
-      if (rp->negate) {
-        switch (rp->relat) {
-        case LE:
-          rp->relat = GE;
-          break;
-        case GE:
-          rp->relat = LE;
-          break;
+      if(rp->SOStype == 0) {
+        if (rp->negate) {
+          switch (rp->relat) {
+          case LE:
+            rp->relat = GE;
+            break;
+          case GE:
+            rp->relat = LE;
+            break;
+          }
+          switch (rp->range_relat) {
+          case LE:
+            rp->range_relat = GE;
+            break;
+          case GE:
+            rp->range_relat = LE;
+            break;
+          }
+          rp->range_value = -rp->range_value;
+          rp->value = -rp->value;
         }
-        switch (rp->range_relat) {
-        case LE:
-          rp->range_relat = GE;
-          break;
-        case GE:
-          rp->range_relat = LE;
-          break;
-        }
-        rp->range_value = -rp->range_value;
-        rp->value = -rp->value;
-      }
 
-      if((rp->range_relat >= 0) && (rp->value == lp->infinite)) {
-        rp->value = rp->range_value;
-        rp->relat = rp->range_relat;
-        rp->range_relat = -1;
+        if((rp->range_relat >= 0) && (rp->value == lp->infinite)) {
+          rp->value = rp->range_value;
+          rp->relat = rp->range_relat;
+          rp->range_relat = -1;
+        }
+        else if((rp->range_relat >= 0) && (rp->value == -lp->infinite)) {
+          rp->value = rp->range_value;
+          rp->relat = rp->range_relat;
+          rp->range_relat = -1;
+        }
+        if ((rp->range_relat >= 0) && (rp->range_value == rp->value)) {
+          rp->relat = EQ;
+          rp->range_relat = EQ;
+        }
+	if(i) {
+          set_constr_type(lp, i, rp->relat);
+	  relat[i] = rp->relat;
+	}
+        set_rh(lp, i, rp->value);
+        if (rp->range_relat >= 0)
+          set_rh_range(lp, i, rp->range_value - rp->value);
       }
-      else if((rp->range_relat >= 0) && (rp->value == -lp->infinite)) {
-        rp->value = rp->range_value;
-        rp->relat = rp->range_relat;
-        rp->range_relat = -1;
+      else {
+        SOSinMatrix = TRUE;
+        if(i)
+          relat[i] = rp->relat;
       }
-      if ((rp->range_relat >= 0) && (rp->range_value == rp->value)) {
-        rp->relat = EQ;
-        rp->range_relat = EQ;
-      }
-      if(i) {
-        set_constr_type(lp, i, rp->relat);
-        relat[i] = rp->relat;
-      }
-      set_rh(lp, i, rp->value);
-      if (rp->range_relat >= 0)
-        /* lp->orig_upbo[i] = my_abs(rp->range_value - rp->value); */
-        /* set_uprange(lp, i, my_abs(rp->range_value - rp->value)); */
-        set_rh_range(lp, i, rp->range_value - rp->value);
     }
     if(rp != NULL) {
       First_rside = rp->next;
@@ -852,9 +953,17 @@ static int readinput(lprec *lp)
   /* start reading the Hash_list structure */
   index = 0;
 
+  if((SOSinMatrix) && (CALLOC(SOSrowdata, 1 + Rows, struct SOSrowdata) == NULL)) {
+    FREE(negateAndSOS);
+    FREE(row);
+    FREE(rowno);
+    return(FALSE);
+  }
+
   if((lp != NULL) &&
      ((MALLOC(row, 1 + Rows, REAL) == NULL) || (MALLOC(rowno, 1 + Rows, int) == NULL))) {
-    FREE(negate);
+    FREE(SOSrowdata);
+    FREE(negateAndSOS);
     FREE(row);
     FREE(rowno);
     return(FALSE);
@@ -867,13 +976,32 @@ static int readinput(lprec *lp)
       count = 0;
       index++;
       cp = coldata[hp->index].firstcol;
+      col = hp->index + 1;
       while(cp != NULL) {
         if(lp != NULL) {
-	  rowno[count] = cp->row;
-	  a = cp->value;
-	  if (negate[cp->row])
-	    a = -a;
-          row[count++] = a;
+          if (negateAndSOS[cp->row] <= 0) {
+            rowno[count] = cp->row;
+  	    a = cp->value;
+  	    if (negateAndSOS[cp->row])
+  	      a = -a;
+            row[count++] = a;
+          }
+	  else {
+	    if (MALLOC(SOSrow, 1, struct SOSrow) == NULL) {
+              FREE(SOSrowdata);
+              FREE(negateAndSOS);
+              FREE(row);
+              FREE(rowno);
+              return(FALSE);
+	    }
+	    if(SOSrowdata[cp->row].SOSrow == NULL)
+	      SOSrowdata[cp->row].name = strdup(get_row_name(lp, cp->row));
+	    SOSrow->next = SOSrowdata[cp->row].SOSrow;
+	    SOSrowdata[cp->row].SOSrow = SOSrow;
+	    SOSrowdata[cp->row].type = negateAndSOS[cp->row];
+	    SOSrow->col = col;
+	    SOSrow->value = cp->value;
+	  }
 	}
 	tcp = cp;
 	/* cp = cp->next; */
@@ -901,10 +1029,14 @@ static int readinput(lprec *lp)
         if(coldata[hp->index].must_be_sec) {
           set_semicont(lp, index, TRUE);
         }
+        if(coldata[hp->index].must_be_free) {
+          set_unbounded(lp, index);
+        }
 
         /* copy name of column variable */
         if (!set_col_name(lp, index, hp->name)) {
-          FREE(negate);
+	  FREE(SOSrowdata);
+          FREE(negateAndSOS);
           FREE(row);
 	  FREE(rowno);
           return(FALSE);
@@ -927,10 +1059,51 @@ static int readinput(lprec *lp)
 
   FREE(coldata);
 
+  if(SOSrowdata != NULL) {
+    struct structSOS *structSOS;
+    struct structSOSvars *SOSvars, *SOSvars1;
+    int SOSweight = 0;
+
+    for(i = 1; i <= Rows; i++) {
+      SOSrow = SOSrowdata[i].SOSrow;
+      if(SOSrow != NULL) {
+	if(MALLOC(structSOS, 1, struct structSOS) == NULL) {
+	  FREE(SOSrowdata);
+          FREE(negateAndSOS);
+          FREE(row);
+	  FREE(rowno);
+          return(FALSE);
+	}
+	structSOS->Nvars = 0;
+	structSOS->type = SOSrowdata[i].type;
+	structSOS->weight = ++SOSweight;
+	structSOS->name = strdup(SOSrowdata[i].name);
+	structSOS->LastSOSvars = NULL;
+	structSOS->next = FirstSOS;
+	FirstSOS = structSOS;
+	SOSvars = NULL;
+        while(SOSrow != NULL) {
+	  SOSvars1 = SOSvars;
+	  MALLOC(SOSvars, 1, struct structSOSvars);
+	  SOSvars->next = SOSvars1;
+	  SOSvars->col = SOSrow->col;
+	  SOSvars->weight = SOSrow->value;
+	  SOSvars->name = NULL;
+	  structSOS->Nvars++;
+          SOSrow1 = SOSrow->next;
+	  FREE(SOSrow);
+	  SOSrow = SOSrow1;
+	}
+	structSOS->SOSvars = SOSvars;
+      }
+    }
+    FREE(SOSrowdata);
+  }
+
   while(FirstSOS != NULL)
   {
     struct structSOSvars *SOSvars, *SOSvars1;
-    int *sosvars, n;
+    int *sosvars, n, col;
     REAL *weights;
     hashelem *hp;
 
@@ -950,9 +1123,15 @@ static int readinput(lprec *lp)
     {
       SOSvars1 = SOSvars;
       SOSvars = SOSvars->next;
-      if((lp != NULL) && ((hp = findhash(SOSvars1->name, lp->colname_hashtab)) != NULL)) {
-        sosvars[n] = hp->index;
-	weights[n++] = SOSvars1->weight;
+      if(lp != NULL) {
+        col = SOSvars1->col;
+	if(col == 0)
+          if((hp = findhash(SOSvars1->name, lp->colname_hashtab)) != NULL)
+	    col = hp->index;
+	if (col) {
+          sosvars[n] = col;
+	  weights[n++] = SOSvars1->weight;
+	}
       }
       FREE(SOSvars1->name);
       FREE(SOSvars1);
@@ -966,6 +1145,20 @@ static int readinput(lprec *lp)
     FREE(LastSOS);
   }
 
+  if(negateAndSOS != NULL) {
+    for(i1 = 0, i = 1; i <= Rows; i++)
+      if(negateAndSOS[i] <= 0)
+        relat[++i1] = relat[i];
+
+#if 01
+    for(i = Rows; i > 0; i--)
+      if(negateAndSOS[i] > 0) {
+        del_constraint(lp, i);
+        Rows--;
+      }
+#endif
+  }
+
   /* the following should be replaced by a call to the MPS print routine MB */
 
 #if 0
@@ -1040,7 +1233,7 @@ static int readinput(lprec *lp)
 
   FREE(row);
   FREE(rowno);
-  FREE(negate);
+  FREE(negateAndSOS);
   return(TRUE);
 } /* readinput */
 
@@ -1112,5 +1305,6 @@ lprec *yacc_read(lprec *lp, int verbose, char *lp_name, int *_lineno, int (*pars
     }
     FREE(relat);
   }
+  null_tmp_store(FALSE);
   return(lp);
 }
